#!/usr/bin/python
#
# Nikola - firewall log sender (a part of www.turris.cz project)
# Copyright (C) 2013 CZ.NIC, z.s.p.o. (http://www.nic.cz/)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#


import base64
import json
import argparse
import os
import random
import socket
import ssl
import subprocess
import sys
import time
import traceback
import zlib

from nikola.filter import filter_records
from nikola.logger import get_logger
from nikola.rpc_wrapper import WrappedServer
from nikola.syslog_parser import parse_syslog
from nikola.tester import test_connect, publish_result


logger = None


def send_test_packet(test_ip):
    logger.debug("Trying to send a testing packet (it should be blocked by the firewall).")
    try:
        test_connect(test_ip)
        logger.error("turris firewall rules might not be active")
    except socket.error as e:
        # sending should failed when the rules are applied
        logger.debug("Failed to send the testing packet (this is expected - blocked by firewall)")
    except Exception as e:
        logger.error("failed to send the testing packet: %s" % e)
        e_type, e_value, e_traceback = sys.exc_info()
        logger.error("Exception traceback: %s" % str(traceback.extract_tb(e_traceback)))


def main():

    # Parse the command line options
    parser = argparse.ArgumentParser(prog="nikola")
    parser.add_argument(
        "-s", "--server-address", dest='server_address', default='https://api.turris.cz/fw',
        type=str, help='set the server address'
    )

    parser.add_argument(
        "-l", "--log-file", dest='syslog_file', default='/var/log/iptables', type=str,
        help='specify the syslog file to be parsed'
    )

    parser.add_argument(
        "-m", "--max", dest='max', default=1000, help='max record count to be sent', type=int,
    )

    parser.add_argument(
        "-f", "--date-format", dest='date_format', default='%Y-%m-%dT%H:%M:%S', type=str,
        help='specify the syslog date format'
    )

    parser.add_argument(
        "-r", "--log-rotate-conf", dest='logrotate_conf', default='/etc/logrotate.d/iptables',
        type=str, help='specify the log rotate config to be triggered'
    )

    parser.add_argument(
        "-d", "--debug", dest='debug', action='store_true', default=False,
        help='use debug output'
    )

    parser.add_argument(
        "-w", "--wan-interfaces", dest='wans', default=None, type=str,
        help='comma separated list of wan interfaces'
    )

    parser.add_argument(
        "-t", "--test-ip", dest='test_ip', default='192.0.2.84', type=str,
        help='the address which is used for rule testing purposes'
    )

    parser.add_argument(
        "-u", "--test-ruleid", dest='test_rule_id', default='00DEB060', type=str,
        help='this rule id is used for a testing purposes'
    )

    parser.add_argument(
        "-j", "--just-test", dest='just_test', action='store_true', default=False,
        help='just try to send a test packet'
    )

    parser.add_argument(
        "-c", "--server-certificate", dest='certificate', default=None,
        type=str, help='server certificate to check')

    parser.add_argument(
        "-C", "--ca-path", dest='ca_path', default='/etc/ssl/turris.pem',
        type=str, help='certificate autohrity path'
    )

    parser.add_argument(
        "-L", "--crl-path", dest='crl_path', default='/etc/ssl/crl.pem',
        type=str, help='certificate autohrity path'
    )

    parser.add_argument(
        "-i", "--ignore-certificate", dest='ignore_certificate', action='store_true',
        default=False, help='don\'t check the server certificate'
    )

    parser.add_argument(
        "-n", "--now", dest='now', action='store_true', default=False,
        help='don\'t use random sleep interval'
    )

    options = parser.parse_args()

    global logger
    logger = get_logger(options.debug)

    syslog_file = options.syslog_file
    max_packet_count = options.max
    syslog_date_format = options.date_format
    logrotate_conf = options.logrotate_conf
    wans = [e.strip() for e in options.wans.split(",") if e.strip()] if options.wans else []
    certificate = options.certificate
    ignore_certificate = options.ignore_certificate
    ca_path = options.ca_path
    crl_path = options.crl_path
    now = options.now
    test_ip = options.test_ip
    test_rule_id = options.test_rule_id
    just_test = options.just_test
    server_address = options.server_address
    if just_test:
        send_test_packet(test_ip)
        sys.exit()

    parsed = []

    logger.debug("Nikola synchronization is starting...")
    logger.info("recognized WAN interfaces: " + ", ".join(wans))

    if not now:
        # sleep for random amount of time up to one minute
        seconds = random.randrange(60)
        logger.debug("Sleeping for %d seconds." % seconds)
        time.sleep(seconds)

    last_time = time.time()

    failed = False
    try:

        # CA and crl validation
        if not ignore_certificate:
            # prepare ssl context here
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            ssl_context.load_verify_locations(ca_path)
            ssl_context.load_verify_locations(crl_path)
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            ssl_context.verify_flags |= ssl.VERIFY_CRL_CHECK_CHAIN
        else:
            ssl_context = None

        server = WrappedServer(server_address, ssl_context=ssl_context)
        server.connect_when_not_connected()

        logger.info("Establishing connection took %f seconds" % (time.time() - last_time))
        last_time = time.time()

        # Skip the certificate validation when the flag is checked or the certificate is not set
        if not ignore_certificate and certificate:

            # Validate the server certificate
            with open(certificate) as f:
                certificate = f.read()

            server.check_certificate(certificate)
            logger.info("Checking server certificates took %f seconds" % (time.time() - last_time))
            last_time = time.time()

        if not wans:
            logger.warning('No WAN device was set. No data will be send to the server!')

        if os.path.exists(syslog_file) and wans:
            # logrotete the logs
            output = subprocess.check_output(
                ('/usr/sbin/logrotate', '-s', '/tmp/logrotate-nikola.state', logrotate_conf, )
            )
            logger.debug(("logrotate output: %s" % output))

            logger.info("Logrotate took %f seconds" % (time.time() - last_time))
            last_time = time.time()

            # Parse syslog
            parsed = parse_syslog("%s.1" % syslog_file, wans, syslog_date_format, logger=logger)

            logger.info("Syslog parsing took %f seconds" % (time.time() - last_time))
            last_time = time.time()

        else:
            # To file to parse means no records
            parsed = []

        init_res = server.init_session()
        logger.debug(init_res)
        logger.info("Session init took %f seconds" % (time.time() - last_time))
        last_time = time.time()

        exclude_remote_regexps = init_res.get('fw_exclude_regexp', [])
        exclude_local_regexps = init_res.get('fw_local_exclude_regexp', [])

        logger.info(("Records parsed: %d" % len(parsed)))
        parsed = filter_records(
            parsed, max_packet_count, exclude_remote_regexps, exclude_local_regexps)
        logger.info(("Records after filtering: %d" % len(parsed)))
        logger.info("Records filtering took %f seconds" % (time.time() - last_time))
        last_time = time.time()

        logger.debug(("First record: %s" % parsed[0]) if parsed else 'No records')
        if max_packet_count < len(parsed):
            msg = "Too many records to send in one batch. (%d/%d)" % (len(parsed),
                                                                      max_packet_count)
            logger.error(msg)
            logger.debug(server.api_turris_cz.report_error(msg))
            logger.info("Sending error took %f seconds" % (time.time() - last_time))
            last_time = time.time()
        else:
            # comporess the logs
            compressed = base64.b64encode(zlib.compress(json.dumps(parsed).encode(), 1))
            logger.info(server.api_turris_cz.firewall.store_logs(compressed))
            logger.info("Sending records took %f seconds" % (time.time() - last_time))
            last_time = time.time()

    except socket.error as e:
        failed = True
        logger.error("socket: %s" % e)
    except Exception as e:
        failed = True
        logger.error("Exception thrown: %s" % e)
        e_type, e_value, e_traceback = sys.exc_info()
        logger.error("Exception traceback: %s" % str(traceback.extract_tb(e_traceback)))

    # publish whether the rules are applied
    try:
        publish_result(parsed, test_rule_id, failed)
    except Exception as e:
        e_type, e_value, e_traceback = sys.exc_info()
        logger.error("error during rule test: %s" % e)
        e_type, e_value, e_traceback = sys.exc_info()
        logger.error("Exception traceback: %s" % str(traceback.extract_tb(e_traceback)))

    # try to send the testing packet
    send_test_packet(test_ip)


if __name__ == '__main__':
    main()
